"""Part 2 of migration to switch recurrence_unit to CharField."""
# pylint: disable=invalid-name, missing-docstring
from django.db import migrations


def convert_recurrence_unit_forward(apps, schema_editor):  # pylint: disable=unused-argument
    """Copy integer-based unit to char-based."""
    PlanCost = apps.get_model('subscriptions', 'PlanCost')

    # Update recurrence unit for all PlanCost instances
    for cost in PlanCost.objects.all():
        old_unit = cost.old_recurrence_unit
        new_unit = str(old_unit)
        cost.recurrence_unit = new_unit
        cost.save()


def convert_recurrence_unit_reverse(apps, schema_editor):  # pylint: disable=unused-argument
    """Copy char-based unit to integer-based."""
    PlanCost = apps.get_model('subscriptions', 'PlanCost')

    # Update recurrence unit for all PlanCost instances
    for cost in PlanCost.objects.all():
        new_unit = cost.recurrence_unit
        old_unit = int(new_unit)
        cost.old_recurrence_unit = old_unit
        cost.save()


class Migration(migrations.Migration):
    dependencies = [
        ('subscriptions', '0004_change_recurrence_unit_type_1'),
    ]

    operations = [
        migrations.RunPython(convert_recurrence_unit_forward, convert_recurrence_unit_reverse),
        migrations.RemoveField('plancost', 'old_recurrence_unit'),
    ]



"""Abstract templates for the Djanog Flexible Subscriptions app."""
from django.views import generic

from subscriptions.conf import SETTINGS

BASE_TEMPLATE = SETTINGS['base_template']


class TemplateView(generic.TemplateView):
    """Extends TemplateView to specify of extensible HTML template.

        Attributes:
            template_extends (str): Path to HTML template that this
                view extends.
    """
    template_extends = BASE_TEMPLATE

    def get_context_data(self, **kwargs):
        """Overriding get_context_data to add additional context."""
        context = super(TemplateView, self).get_context_data(**kwargs)

        # Provides the base template to extend from
        context['template_extends'] = self.template_extends

        return context


class ListView(generic.ListView):
    """Extends ListView to specify of extensible HTML template

        Attributes:
            template_extends (str): Path to HTML template that this
                view extends.
    """
    template_extends = BASE_TEMPLATE

    def get_context_data(self, *, object_list=None, **kwargs):  # pylint: disable=unused-argument
        """Overriding get_context_data to add additional context."""
        context = super(ListView, self).get_context_data(**kwargs)

        # Provides the base template to extend from
        context['template_extends'] = self.template_extends

        return context


class DetailView(generic.DetailView):
    """Extends DetailView to specify of extensible HTML template

        Attributes:
            template_extends (str): Path to HTML template that this
                view extends.
    """
    template_extends = BASE_TEMPLATE

    def get_context_data(self, **kwargs):
        """Overriding get_context_data to add additional context."""
        context = super(DetailView, self).get_context_data(**kwargs)

        # Provides the base template to extend from
        context['template_extends'] = self.template_extends

        return context


class CreateView(generic.CreateView):
    """Extends CreateView to specify of extensible HTML template

        Attributes:
            template_extends (str): Path to HTML template that this
                view extends.
    """
    template_extends = BASE_TEMPLATE

    def get_context_data(self, **kwargs):
        """Overriding get_context_data to add additional context."""
        context = super(CreateView, self).get_context_data(**kwargs)

        # Provides the base template to extend from
        context['template_extends'] = self.template_extends

        return context


class UpdateView(generic.UpdateView):
    """Extends UpdateView to specify of extensible HTML template

        Attributes:
            template_extends (str): Path to HTML template that this
                view extends.
    """
    template_extends = BASE_TEMPLATE

    def get_context_data(self, **kwargs):
        """Overriding get_context_data to add additional context."""
        context = super(UpdateView, self).get_context_data(**kwargs)

        # Provides the base template to extend from
        context['template_extends'] = self.template_extends

        return context


class DeleteView(generic.DeleteView):
    """Extends DeleteView to specify of extensible HTML template

        Attributes:
            template_extends (str): Path to HTML template that this
                view extends.
    """
    template_extends = BASE_TEMPLATE

    def get_context_data(self, **kwargs):
        """Overriding get_context_data to add additional context."""
        context = super(DeleteView, self).get_context_data(**kwargs)

        # Provides the base template to extend from
        context['template_extends'] = self.template_extends

        return context



"""Initial model creation."""
# pylint: disable=missing-docstring, invalid-name

import uuid

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('auth', '0008_alter_user_username_max_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='PlanCost',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'recurrence_period',
                    models.PositiveSmallIntegerField(
                        default=1,
                        help_text='how often the plan is billed (per recurrence unit)',
                        validators=[django.core.validators.MinValueValidator(1)],
                    ),
                ),
                (
                    'recurrence_unit',
                    models.PositiveIntegerField(
                        default=6,
                        help_text='the unit of measurement for the recurrence period',
                        validators=[django.core.validators.MaxValueValidator(7)],
                    ),
                ),
                (
                    'cost',
                    models.DecimalField(
                        blank=True,
                        decimal_places=4,
                        help_text='the cost per recurrence of the plan',
                        max_digits=19,
                        null=True,
                    ),
                ),
            ],
            options={
                'ordering': ('recurrence_unit', 'recurrence_period', 'cost',),
            }
        ),
        migrations.CreateModel(
            name='PlanTag',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'tag',
                    models.CharField(
                        help_text='the tag name',
                        max_length=64,
                        unique=True,
                    ),
                ),
            ],
            options={
                'ordering': ('tag',),
            },
        ),
        migrations.CreateModel(
            name='SubscriptionPlan',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'plan_name',
                    models.CharField(
                        help_text='the name of the subscription plan',
                        max_length=128,
                    ),
                ),
                (
                    'plan_description',
                    models.CharField(
                        blank=True,
                        help_text='a description of the subscription plan',
                        max_length=512,
                        null=True,
                    ),
                ),
                (
                    'grace_period',
                    models.PositiveIntegerField(
                        default=0,
                        help_text='how many days after the subscription ends before the subscription expires',
                    ),
                ),
                (
                    'group',
                    models.ForeignKey(
                        blank=True,
                        help_text='the Django auth group for this plan',
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name='plans',
                        to='auth.Group',
                    ),
                ),
                (
                    'tags',
                    models.ManyToManyField(
                        blank=True,
                        help_text='any tags associated with this plan',
                        related_name='plans',
                        to='subscriptions.PlanTag',
                    ),
                ),
            ],
            options={
                'ordering': ('plan_name',),
                'permissions': (('subscriptions', 'Can interact with subscription details'),),
            },
        ),
        migrations.CreateModel(
            name='SubscriptionTransaction',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'date_transaction',
                    models.DateTimeField(
                        help_text='the datetime the transaction was billed',
                        verbose_name='transaction date',
                    ),
                ),
                (
                    'amount',
                    models.DecimalField(
                        blank=True,
                        decimal_places=4,
                        help_text='how much was billed for the user',
                        max_digits=19, null=True,
                    ),
                ),
                (
                    'subscription',
                    models.ForeignKey(
                        help_text='the plan costs that were billed',
                        null=True, on_delete=django.db.models.deletion.SET_NULL,
                        related_name='transactions',
                        to='subscriptions.PlanCost',
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        help_text='the user that this subscription was billed for',
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name='subscription_transactions',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                'ordering': ('date_transaction', 'user',),
            },
        ),
        migrations.CreateModel(
            name='UserSubscription',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'date_billing_start',
                    models.DateTimeField(
                        blank=True,
                        help_text='the date to start billing this subscription',
                        null=True,
                        verbose_name='billing start date',
                    ),
                ),
                (
                    'date_billing_end',
                    models.DateTimeField(
                        blank=True,
                        help_text='the date to finish billing this subscription',
                        null=True,
                        verbose_name='billing start end',
                    ),
                ),
                (
                    'date_billing_last',
                    models.DateTimeField(
                        blank=True,
                        help_text='the last date this plan was billed',
                        null=True,
                        verbose_name='last billing date',
                    ),
                ),
                (
                    'date_billing_next',
                    models.DateTimeField(
                        blank=True,
                        help_text='the next date billing is due',
                        null=True,
                        verbose_name='next start date',
                    ),
                ),
                (
                    'active',
                    models.BooleanField(
                        default=True,
                        help_text='whether this subscription is active or not',
                    ),
                ),
                (
                    'cancelled',
                    models.BooleanField(
                        default=False,
                        help_text='whether this subscription is cancelled or not',
                    ),
                ),
                (
                    'subscription',
                    models.ForeignKey(
                        help_text='the plan costs and billing frequency for this user',
                        null=True, on_delete=django.db.models.deletion.CASCADE,
                        related_name='subscriptions',
                        to='subscriptions.PlanCost',
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        help_text='the user this subscription applies to',
                        null=True, on_delete=django.db.models.deletion.CASCADE,
                        related_name='subscriptions',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                'ordering': ('user', 'date_billing_start',),
            },
        ),
        migrations.AddField(
            model_name='plancost',
            name='plan',
            field=models.ForeignKey(
                help_text='the subscription plan for these cost details',
                on_delete=django.db.models.deletion.CASCADE,
                related_name='costs',
                to='subscriptions.SubscriptionPlan',
            ),
        ),
    ]



"""Django management command to process subscriptions via task runner."""
import importlib

from django.core.management.base import BaseCommand

from subscriptions.conf import SETTINGS


class Command(BaseCommand):
    """Django management command to process subscriptions via task runner."""
    help = 'Processes all subscriptions to handle renewal and expiries.'

    def handle(self, *args, **options):
        """Runs Manager methods required to process subscriptions."""
        Manager = getattr(  # pylint: disable=invalid-name
            importlib.import_module(SETTINGS['management_manager']['module']),
            SETTINGS['management_manager']['class']
        )
        manager = Manager()

        self.stdout.write('Processing subscriptions... ', ending='')
        manager.process_subscriptions()
        self.stdout.write('Complete!')



# pylint: disable=missing-docstring, invalid-name
import sys
import re
import warnings
import django

__version__ = '0.15.0'

# Provide DeprecationWarning for older Python versions
# Have to use sys.version while supporting Python 3.5 to enable testing
# Once Python 3.5 is dropped can switch to version_info & compare tuples
if re.match(r'^3\.5', sys.version):
    warnings.warn(
        (
            'django-flexible-subscription will stop supporting Python 3.5 '
            'once it reaches end-of-life (approximately September 2020). '
            'Ensure you have updated your Python version by then.'
        ),
        DeprecationWarning
    )
# Provide DeprecationWarning for older Django versions
# if '3.0' in django.__version__:
#     warnings.warn(
#         (
#             'django-flexible-subscription will stop supporting Django 3.0 LTS '
#             'once it reaches end-of-life (approximately April 2021). '
#             'Ensure you have updated your Django version by then.'
#         ),
#         DeprecationWarning
#     )

# Django configuration details
default_app_config = 'subscriptions.apps.FlexibleSubscriptionsConfig'



"""Functions for general package configuration."""
import warnings

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from subscriptions.currency import Currency, CURRENCY


def string_to_module_and_class(string):
    """Breaks a string to a module and class name component."""
    components = string.split('.')
    component_class = components.pop()
    component_module = '.'.join(components)

    return {
        'module': component_module,
        'class': component_class,
    }


def validate_currency_settings(currency_locale):
    """Validates provided currency settings.

        Parameters
            currency_locale (str or dict): a currency locale string or
                a dictionary defining custom currency formating
                conventions.

        Raises:
            ImproperlyConfigured: specified string currency_locale not
                support.
            TypeError: invalid parameter type provided.
    """
    # STRING VALIDATION
    # ------------------------------------------------------------------------
    if isinstance(currency_locale, str):
        # Confirm that the provided locale is supported
        if currency_locale.lower() not in CURRENCY:
            raise ImproperlyConfigured(
                '{} is not a support DFS_CURRENCY_LOCALE value.'.format(currency_locale)
            )
    elif isinstance(currency_locale, dict):
        # Placeholder for any future specific dictionary validation
        pass
    else:
        raise TypeError(
            'Invalid DFS_CURRENCY_LOCALE type: {}. Must be str or dict.'.format(
                type(currency_locale)
            )
        )


def determine_currency_settings():
    """Determines details for Currency handling.

        Validates the provided currency locale setting and then returns
        a Currency object.

        Returns:
            obj: a Currency object for the provided setting.
    """
    # Get the proper setting attribute name
    # This block can be removed when DFS_CURRENCY_LOCALE has been
    # removed
    if hasattr(settings, 'DFS_CURRENCY'):
        currency_setting = 'DFS_CURRENCY'
    elif hasattr(settings, 'DFS_CURRENCY_LOCALE'):
        currency_setting = 'DFS_CURRENCY_LOCALE'

        deprecation_warning = (
            'DFS_CURRENCY_LOCALE is deprecated and has been replaced by '
            'DFS_CURRENCY. DFS_CURRENCY_LOCALE will be removed in a '
            'future version of django-flexible-subscription.'
        )
        warnings.warn(deprecation_warning, DeprecationWarning)
    else:
        currency_setting = 'DFS_CURRENCY'

    # Get the value for the currency
    currency_value = getattr(settings, currency_setting, 'en_us')

    # Validate currency locale setting
    validate_currency_settings(currency_value)

    # Return the Currency object
    return Currency(currency_value)


def compile_settings():
    """Compiles and validates all package settings and defaults.

        Provides basic checks to ensure required settings are declared
        and applies defaults for all missing settings.

        Returns:
            dict: All possible Django Flexible Subscriptions settings.
    """
    # ADMIN SETTINGS
    # -------------------------------------------------------------------------
    enable_admin = getattr(settings, 'DFS_ENABLE_ADMIN', False)

    # CURRENCY SETTINGS
    # -------------------------------------------------------------------------
    currency = determine_currency_settings()

    # TEMPLATE & VIEW SETTINGS
    # -------------------------------------------------------------------------
    base_template = getattr(
        settings, 'DFS_BASE_TEMPLATE', 'subscriptions/base.html'
    )

    # Get module and class for SubscribeView
    subscribe_view_path = getattr(
        settings, 'DFS_SUBSCRIBE_VIEW', 'subscriptions.views.SubscribeView'
    )
    subscribe_view = string_to_module_and_class(subscribe_view_path)

    # MANAGEMENT COMMANDS SETTINGS
    # ------------------------------------------------------------------------
    # Get module and class for the Management Command Manager class
    manager_object = getattr(
        settings,
        'DFS_MANAGER_CLASS',
        'subscriptions.management.commands._manager.Manager',
    )
    management_manager = string_to_module_and_class(manager_object)

    return {
        'enable_admin': enable_admin,
        'currency': currency,
        'base_template': base_template,
        'subscribe_view': subscribe_view,
        'management_manager': management_manager,
    }


SETTINGS = compile_settings()



"""Forms for Django Flexible Subscriptions."""
# pylint: disable=invalid-name
from django import forms
from django.core import validators
from django.forms import ModelForm
from django.utils import timezone

from subscriptions.conf import SETTINGS
from subscriptions.models import SubscriptionPlan, PlanCost


def assemble_cc_years():
    """Creates a list of the next 60 years."""
    cc_years = []
    now = timezone.now()

    for year in range(now.year, now.year + 60):
        cc_years.append((year, year))

    return cc_years


class SubscriptionPlanForm(ModelForm):
    """Model Form for SubscriptionPlan model."""
    class Meta:
        model = SubscriptionPlan
        fields = [
            'plan_name', 'plan_description', 'group', 'tags', 'grace_period',
        ]


class PlanCostForm(ModelForm):
    """Form to use with inlineformset_factory and SubscriptionPlanForm."""
    class Meta:
        model = PlanCost
        fields = ['recurrence_period', 'recurrence_unit', 'cost']


class PaymentForm(forms.Form):
    """Form to collect details required for payment billing."""
    CC_MONTHS = (
        ('1', '01 - January'),
        ('2', '02 - February'),
        ('3', '03 - March'),
        ('4', '04 - April'),
        ('5', '05 - May'),
        ('6', '06 - June'),
        ('7', '07 - July'),
        ('8', '08 - August'),
        ('9', '09 - September'),
        ('10', '10 - October'),
        ('11', '11 - November'),
        ('12', '12 - December'),
    )
    CC_YEARS = assemble_cc_years()

    cardholder_name = forms.CharField(
        label='Cardholder name',
        max_length=255,
        min_length=1,
    )
    card_number = forms.CharField(
        label='Card number',
        max_length=19,
        min_length=13,
        validators=[validators.RegexValidator(
            r'^\d{13,19}$',
            message='Invalid credit card number',
        )]
    )
    card_expiry_month = forms.ChoiceField(
        choices=CC_MONTHS,
        label='Card expiry (month)',
    )
    card_expiry_year = forms.ChoiceField(
        choices=CC_YEARS,
        label='Card expiry (year)',
    )
    card_cvv = forms.CharField(
        label='Card CVV',
        max_length=4,
        min_length=3,
        validators=[validators.RegexValidator(
            r'^\d{3,4}$',
            message='Invalid CVV2 number',
        )]
    )
    address_title = forms.CharField(
        label='Title',
        max_length=32,
        required=False,
    )
    address_name = forms.CharField(
        label='Name',
        max_length=128,
    )
    address_line_1 = forms.CharField(
        label='Line 1',
        max_length=256,
    )
    address_line_2 = forms.CharField(
        label='Line 2',
        max_length=256,
        required=False,
    )
    address_line_3 = forms.CharField(
        label='Line 3',
        max_length=256,
        required=False,
    )
    address_city = forms.CharField(
        label='City',
        max_length=128,
        min_length=1,
    )
    address_province = forms.CharField(
        label='Province/State',
        max_length=128,
        min_length=1,
    )
    address_postcode = forms.CharField(
        label='Postcode',
        max_length=16,
        required=False,
    )
    address_country = forms.CharField(
        label='Country',
        max_length=128,
        min_length=1,
    )


class SubscriptionPlanCostForm(forms.Form):
    """Form to handle choosing a subscription plan for payment."""
    plan_cost = forms.UUIDField(
        label='',
        widget=forms.RadioSelect(),
    )

    def __init__(self, *args, **kwargs):
        """Overrides the plan_cost widget with available selections.

            For a provided subscription plan, provides a widget that
            lists all possible plan costs for selection.

            Keyword Arguments:
                subscription_plan (obj): A SubscriptionPlan instance.
        """
        costs = kwargs.pop('subscription_plan').costs.all()
        PLAN_COST_CHOICES = []

        for cost in costs:
            radio_text = '{} {}'.format(
                SETTINGS['currency'].format_currency(cost.cost),
                cost.display_billing_frequency_text
            )
            PLAN_COST_CHOICES.append((cost.id, radio_text))

        super(SubscriptionPlanCostForm, self).__init__(*args, **kwargs)

        # Update the radio widget with proper choices
        self.fields['plan_cost'].widget.choices = PLAN_COST_CHOICES

        # Set the last value as the default
        self.fields['plan_cost'].initial = [PLAN_COST_CHOICES[-1][0]]

    def clean_plan_cost(self):
        """Validates that UUID is valid and returns model instance."""
        try:
            data = PlanCost.objects.get(id=self.cleaned_data['plan_cost'])
        except PlanCost.DoesNotExist:
            raise forms.ValidationError('Invalid plan cost submitted.')

        return data



"""paths for the Flexible Subscriptions app."""
# pylint: disable=line-too-long
import importlib

from django.urls import path

from subscriptions import views
from subscriptions.conf import SETTINGS


# Retrieve the proper subscribe view
SubscribeView = getattr(  # pylint: disable=invalid-name
    importlib.import_module(SETTINGS['subscribe_view']['module']),
    SETTINGS['subscribe_view']['class']
)


urlpatterns = [
    path(
        'subscribe/',
        views.SubscribeList.as_view(),
        name='dfs_subscribe_list',
    ),
    path(
        'subscribe/add/',
        SubscribeView.as_view(),
        name='dfs_subscribe_add',
    ),
    path(
        'subscribe/thank-you/<uuid:transaction_id>/',
        views.SubscribeThankYouView.as_view(),
        name='dfs_subscribe_thank_you',
    ),
    path(
        'subscribe/cancel/<uuid:subscription_id>/',
        views.SubscribeCancelView.as_view(),
        name='dfs_subscribe_cancel',
    ),
    path(
        'subscriptions/',
        views.SubscribeUserList.as_view(),
        name='dfs_subscribe_user_list',
    ),
    path(
        'dfs/tags/',
        views.TagListView.as_view(),
        name='dfs_tag_list',
    ),
    path(
        'dfs/tags/create/',
        views.TagCreateView.as_view(),
        name='dfs_tag_create',
    ),
    path(
        'dfs/tags/<int:tag_id>/',
        views.TagUpdateView.as_view(),
        name='dfs_tag_update',
    ),
    path(
        'dfs/tags/<int:tag_id>/delete/',
        views.TagDeleteView.as_view(),
        name='dfs_tag_delete',
    ),
    path(
        'dfs/plans/',
        views.PlanListView.as_view(),
        name='dfs_plan_list',
    ),
    path(
        'dfs/plans/create/',
        views.PlanCreateView.as_view(),
        name='dfs_plan_create',
    ),
    path(
        'dfs/plans/<uuid:plan_id>/',
        views.PlanUpdateView.as_view(),
        name='dfs_plan_update',
    ),
    path(
        'dfs/plans/<uuid:plan_id>/delete/',
        views.PlanDeleteView.as_view(),
        name='dfs_plan_delete',
    ),
    path(
        'dfs/plan-lists/',
        views.PlanListListView.as_view(),
        name='dfs_plan_list_list',
    ),
    path(
        'dfs/plan-lists/create/',
        views.PlanListCreateView.as_view(),
        name='dfs_plan_list_create',
    ),
    path(
        'dfs/plan-lists/<int:plan_list_id>/',
        views.PlanListUpdateView.as_view(),
        name='dfs_plan_list_update',
    ),
    path(
        'dfs/plan-lists/<int:plan_list_id>/delete/',
        views.PlanListDeleteView.as_view(),
        name='dfs_plan_list_delete',
    ),
    path(
        'dfs/plan-lists/<int:plan_list_id>/details/',
        views.PlanListDetailListView.as_view(),
        name='dfs_plan_list_detail_list',
    ),
    path(
        'dfs/plan-lists/<int:plan_list_id>/details/create/',
        views.PlanListDetailCreateView.as_view(),
        name='dfs_plan_list_detail_create',
    ),
    path(
        'dfs/plan-lists/<int:plan_list_id>/details/<int:plan_list_detail_id>/',
        views.PlanListDetailUpdateView.as_view(),
        name='dfs_plan_list_detail_update',
    ),
    path(
        'dfs/plan-lists/<int:plan_list_id>/details/<int:plan_list_detail_id>/delete/',
        views.PlanListDetailDeleteView.as_view(),
        name='dfs_plan_list_detail_delete',
    ),
    path(
        'dfs/subscriptions/',
        views.SubscriptionListView.as_view(),
        name='dfs_subscription_list',
    ),
    path(
        'dfs/subscriptions/create/',
        views.SubscriptionCreateView.as_view(),
        name='dfs_subscription_create',
    ),
    path(
        'dfs/subscriptions/<uuid:subscription_id>/',
        views.SubscriptionUpdateView.as_view(),
        name='dfs_subscription_update',
    ),
    path(
        'dfs/subscriptions/<uuid:subscription_id>/delete/',
        views.SubscriptionDeleteView.as_view(),
        name='dfs_subscription_delete',
    ),
    path(
        'dfs/transactions/',
        views.TransactionListView.as_view(),
        name='dfs_transaction_list',
    ),
    path(
        'dfs/transactions/<uuid:transaction_id>/',
        views.TransactionDetailView.as_view(),
        name='dfs_transaction_detail',
    ),
    path(
        'dfs/',
        views.DashboardView.as_view(),
        name='dfs_dashboard',
    ),
]



"""Template filters for Django Flexible Subscriptions."""
from django import template

from subscriptions.conf import SETTINGS


register = template.Library()


@register.filter(name='currency')
def currency(value):
    """Displays value as a currency based on the provided settings."""
    return SETTINGS['currency'].format_currency(value)



"""Views for the Flexible Subscriptions app."""
# pylint: disable=too-many-lines, no-self-use
from copy import copy

from django.contrib import messages
from django.contrib.auth import get_user_model
from django.contrib.auth.mixins import (
    LoginRequiredMixin, PermissionRequiredMixin
)
from django.contrib.messages.views import SuccessMessageMixin
from django.forms import HiddenInput
from django.forms.models import inlineformset_factory
from django.http import HttpResponseRedirect
from django.http.response import HttpResponseNotAllowed, HttpResponseNotFound
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse
from django.urls import reverse_lazy
from django.utils import timezone

from subscriptions import models, forms, abstract


# Dashboard View
# -----------------------------------------------------------------------------
class DashboardView(PermissionRequiredMixin, abstract.TemplateView):
    """Dashboard view to manage subscription details."""
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    template_name = 'subscriptions/dashboard.html'


# Tag Views
# -----------------------------------------------------------------------------
class TagListView(PermissionRequiredMixin, abstract.ListView):
    """List of all tags."""
    model = models.PlanTag
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'tags'
    template_name = 'subscriptions/tag_list.html'


class TagCreateView(
        PermissionRequiredMixin, SuccessMessageMixin, abstract.CreateView
):
    """View to create a new tag."""
    model = models.PlanTag
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'tag'
    fields = ['tag']
    success_message = 'Tag successfully added'
    success_url = reverse_lazy('dfs_tag_list')
    template_name = 'subscriptions/tag_create.html'


class TagUpdateView(
        PermissionRequiredMixin, SuccessMessageMixin, abstract.UpdateView
):
    """View to update the details of a tag."""
    model = models.PlanTag
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'tag'
    fields = ['tag']
    success_message = 'Tag successfully updated'
    success_url = reverse_lazy('dfs_tag_list')
    pk_url_kwarg = 'tag_id'
    template_name = 'subscriptions/tag_update.html'


class TagDeleteView(PermissionRequiredMixin, abstract.DeleteView):
    """View to delete a tag.

        View is extended to handle additional attributes noted below.

        Attributes:
            success_message (str): Message to display on successful deletion.
            success_url (str): URL to redirect to on successful deletion.
    """
    model = models.PlanTag
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'tag'
    pk_url_kwarg = 'tag_id'
    success_message = 'Tag successfully deleted'
    success_url = reverse_lazy('dfs_tag_list')
    template_name = 'subscriptions/tag_delete.html'

    def delete(self, request, *args, **kwargs):
        """Override delete to allow success message to be added."""
        messages.success(self.request, self.success_message)
        return super(TagDeleteView, self).delete(request, *args, **kwargs)


# Subscription Plan Views
# -----------------------------------------------------------------------------
class PlanListView(PermissionRequiredMixin, abstract.ListView):
    """List of all subscription plans."""
    model = models.SubscriptionPlan
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'plans'
    template_name = 'subscriptions/plan_list.html'


class PlanCreateView(PermissionRequiredMixin, abstract.CreateView):
    """View to create a new subscription plan.

        View is extended to handle additional attributes noted below.

        Attributes:
            success_message (str): Message to display on successful creation.
            success_url (str): URL to redirect to on successful creation.
    """
    # pylint: disable=arguments-differ, attribute-defined-outside-init
    model = models.SubscriptionPlan
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'plan'
    form_class = forms.SubscriptionPlanForm
    success_message = 'Subscription plan successfully added'
    success_url = reverse_lazy('dfs_plan_list')
    template_name = 'subscriptions/plan_create.html'

    def get(self, request, *args, **kwargs):
        """Overriding get method to handle inline formset."""
        # Setup the formset for PlanCost
        PlanCostFormSet = inlineformset_factory(  # pylint: disable=invalid-name
            parent_model=models.SubscriptionPlan,
            model=models.PlanCost,
            form=forms.PlanCostForm,
            can_delete=False,
            extra=1,
        )

        self.object = None
        form = self.get_form(self.get_form_class())
        cost_forms = PlanCostFormSet()

        return self.render_to_response(
            self.get_context_data(
                form=form,
                cost_forms=cost_forms,
            )
        )

    def post(self, request, *args, **kwargs):
        """Overriding post method to handle inline formsets."""
        # Setup the formset for PlanCost
        PlanCostFormSet = inlineformset_factory(  # pylint: disable=invalid-name
            parent_model=models.SubscriptionPlan,
            model=models.PlanCost,
            form=forms.PlanCostForm,
            can_delete=False,
            extra=1,
        )

        self.object = None
        form = self.get_form(self.get_form_class())
        cost_forms = PlanCostFormSet(self.request.POST)

        if form.is_valid() and cost_forms.is_valid():
            return self.form_valid(form, cost_forms)

        return self.form_invalid(form, cost_forms)

    def form_valid(self, form, cost_forms):
        """Handles processing of valid forms.

            Parameters:
                form (obj): Parent SubscriptionPlanForm instance to
                    process.
                cost_forms (obj): PlanCostFormSet instance to process.

            Returns:
                obj: HttpResponseRedirect object to ``success_url``.
        """
        # Save the models
        self.object = form.save()
        cost_forms.instance = self.object
        cost_forms.save()

        # Generate the success message
        messages.success(self.request, self.success_message)

        return HttpResponseRedirect(self.get_success_url())

    def form_invalid(self, form, cost_forms):
        """Handles re-rendering invalid forms with errors.

            Parameters:
                form (obj): Parent SubscriptionPlanForm instance to
                    return.
                cost_forms (obj): PlanCostFormSet instance to return.

            Returns:
                obj: Renders original page with form content."""
        return self.render_to_response(
            self.get_context_data(
                form=form,
                cost_forms=cost_forms,
            )
        )


class PlanUpdateView(PermissionRequiredMixin, abstract.UpdateView):
    """View to update a subscription plan.

        View is extended to handle additional attributes noted below.

        Attributes:
            success_message (str): Message to display on successful creation.
            success_url (str): URL to redirect to on successful creation.

    """
    # pylint: disable=arguments-differ, attribute-defined-outside-init
    model = models.SubscriptionPlan
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'plan'
    form_class = forms.SubscriptionPlanForm
    pk_url_kwarg = 'plan_id'
    success_message = 'Subscription plan successfully updated'
    success_url = reverse_lazy('dfs_plan_list')
    template_name = 'subscriptions/plan_update.html'

    def get(self, request, *args, **kwargs):
        """Overriding get method to handle inline formset."""
        # Setup the formset for PlanCost
        PlanCostFormSet = inlineformset_factory(  # pylint: disable=invalid-name
            parent_model=models.SubscriptionPlan,
            model=models.PlanCost,
            form=forms.PlanCostForm,
            can_delete=True,
            extra=1,
        )

        self.object = self.get_object()
        form = self.get_form(self.get_form_class())
        cost_forms = PlanCostFormSet(instance=self.object)

        return self.render_to_response(
            self.get_context_data(
                form=form,
                cost_forms=cost_forms,
            )
        )

    def post(self, request, *args, **kwargs):
        """Overriding post method to handle inline formsets."""
        # Setup the formset for PlanCost
        PlanCostFormSet = inlineformset_factory(  # pylint: disable=invalid-name
            parent_model=models.SubscriptionPlan,
            model=models.PlanCost,
            form=forms.PlanCostForm,
            can_delete=True,
            extra=1,
        )

        self.object = self.get_object()
        form = self.get_form(self.get_form_class())
        cost_forms = PlanCostFormSet(self.request.POST, instance=self.object)

        if form.is_valid() and cost_forms.is_valid():
            return self.form_valid(form, cost_forms)

        return self.form_invalid(form, cost_forms)

    def form_valid(self, form, cost_forms):
        """Handles processing of valid forms.

            Parameters:
                form (obj): Parent SubscriptionPlanForm instance to
                    process.
                cost_forms (obj): PlanCostFormSet instance to process.

            Returns:
                obj: HttpResponseRedirect object to ``success_url``.
        """
        # Save the models
        self.object = form.save()
        cost_forms.instance = self.object
        cost_forms.save()

        # Generate the success message
        messages.success(self.request, self.success_message)

        return HttpResponseRedirect(self.get_success_url())

    def form_invalid(self, form, cost_forms):
        """Handles re-rendering invalid forms with errors.

            Parameters:
                form (obj): Parent SubscriptionPlanForm instance to
                    return.
                cost_forms (obj): PlanCostFormSet instance to return.

            Returns:
                obj: Renders original page with form content."""
        return self.render_to_response(
            self.get_context_data(
                form=form,
                cost_forms=cost_forms,
            )
        )


class PlanDeleteView(PermissionRequiredMixin, abstract.DeleteView):
    """View to delete a subscription plan.

        View is extended to handle additional attributes noted below.

        Attributes:
            success_message (str): Message to display on successful creation.
            success_url (str): URL to redirect to on successful creation.
    """
    model = models.SubscriptionPlan
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'plan'
    pk_url_kwarg = 'plan_id'
    success_message = 'Subscription plan successfully deleted'
    success_url = reverse_lazy('dfs_plan_list')
    template_name = 'subscriptions/plan_delete.html'

    def delete(self, request, *args, **kwargs):
        """Override delete to allow success message to be added."""
        messages.success(self.request, self.success_message)
        return super(PlanDeleteView, self).delete(request, *args, **kwargs)


# User Subscription Views
# -----------------------------------------------------------------------------.
class SubscriptionListView(PermissionRequiredMixin, abstract.ListView):
    """List of all subscriptions for the users"""
    model = get_user_model()
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'users'
    queryset = model.objects.all().exclude(subscriptions=None).order_by('username')
    paginate_by = 100
    template_name = 'subscriptions/subscription_list.html'


class SubscriptionCreateView(
        PermissionRequiredMixin, SuccessMessageMixin, abstract.CreateView
):
    """View to create a new user subscription."""
    model = models.UserSubscription
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'subscription'
    fields = ['user', 'subscription', 'date_billing_start', 'date_billing_end']
    success_message = 'User subscription successfully added'
    success_url = reverse_lazy('dfs_subscription_list')
    template_name = 'subscriptions/subscription_create.html'


class SubscriptionUpdateView(
        PermissionRequiredMixin, SuccessMessageMixin, abstract.UpdateView
):
    """View to update the details of a user subscription."""
    model = models.UserSubscription
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'subscription'
    fields = [
        'subscription', 'date_billing_start', 'date_billing_end',
        'date_billing_last', 'date_billing_next', 'active', 'cancelled'
    ]
    pk_url_kwarg = 'subscription_id'
    success_message = 'User subscription successfully updated'
    success_url = reverse_lazy('dfs_subscription_list')
    template_name = 'subscriptions/subscription_update.html'


class SubscriptionDeleteView(PermissionRequiredMixin, abstract.DeleteView):
    """View to delete a user subscription.

        View is extended to handle additional attributes noted below.

        Attributes:
            success_message (str): Message to display on successful creation.
            success_url (str): URL to redirect to on successful creation.
    """
    model = models.UserSubscription
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'subscription'
    pk_url_kwarg = 'subscription_id'
    success_message = 'User subscription successfully deleted'
    success_url = reverse_lazy('dfs_subscription_list')
    template_name = 'subscriptions/subscription_delete.html'

    def delete(self, request, *args, **kwargs):
        """Override delete to allow success message to be added."""
        messages.success(self.request, self.success_message)
        return super(SubscriptionDeleteView, self).delete(
            request, *args, **kwargs
        )


# Subscription Transaction Views
# -----------------------------------------------------------------------------
class TransactionListView(PermissionRequiredMixin, abstract.ListView):
    """List of all subscription payment transactions."""
    model = models.SubscriptionTransaction
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'transactions'
    paginate_by = 50
    template_name = 'subscriptions/transaction_list.html'


class TransactionDetailView(PermissionRequiredMixin, abstract.DetailView):
    """Shows details of a specific subscription payment transaction."""
    model = models.SubscriptionTransaction
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'transaction'
    pk_url_kwarg = 'transaction_id'
    template_name = 'subscriptions/transaction_detail.html'


# PlanList Views
# -----------------------------------------------------------------------------
class PlanListListView(PermissionRequiredMixin, abstract.ListView):
    """List of plan lists."""
    model = models.PlanList
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'plan_lists'
    template_name = 'subscriptions/plan_list_list.html'


class PlanListCreateView(
        PermissionRequiredMixin, SuccessMessageMixin, abstract.CreateView
):
    """View to create a new plan list."""
    model = models.PlanList
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'plan_list'
    fields = ['title', 'subtitle', 'header', 'footer', 'active']
    success_message = 'Plan list successfully added'
    success_url = reverse_lazy('dfs_plan_list_list')
    template_name = 'subscriptions/plan_list_create.html'


class PlanListUpdateView(
        PermissionRequiredMixin, SuccessMessageMixin, abstract.UpdateView
):
    """View to update the details of a plan list."""
    model = models.PlanList
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'plan_list'
    fields = ['title', 'subtitle', 'header', 'footer', 'active']
    success_message = 'Plan list successfully updated'
    success_url = reverse_lazy('dfs_plan_list_list')
    pk_url_kwarg = 'plan_list_id'
    template_name = 'subscriptions/plan_list_update.html'


class PlanListDeleteView(PermissionRequiredMixin, abstract.DeleteView):
    """View to delete a plan list.

        View is extended to handle additional attributes noted below.

        Attributes:
            success_message (str): Message to display on successful deletion.
            success_url (str): URL to redirect to on successful deletion.
    """
    model = models.PlanList
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'plan_list'
    pk_url_kwarg = 'plan_list_id'
    success_message = 'Plan list successfully deleted'
    success_url = reverse_lazy('dfs_plan_list_list')
    template_name = 'subscriptions/plan_list_delete.html'

    def delete(self, request, *args, **kwargs):
        """Override delete to allow success message to be added."""
        messages.success(self.request, self.success_message)
        return super(PlanListDeleteView, self).delete(request, *args, **kwargs)


# PlanListDetail Views
# -----------------------------------------------------------------------------
class PlanListDetailListView(PermissionRequiredMixin, abstract.DetailView):
    """List of plan lists."""
    model = models.PlanList
    pk_url_kwarg = 'plan_list_id'
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'plan_list'
    template_name = 'subscriptions/plan_list_detail_list.html'


class PlanListDetailCreateView(
        PermissionRequiredMixin, SuccessMessageMixin, abstract.CreateView
):
    """View to create a new plan list."""
    model = models.PlanListDetail
    fields = [
        'plan', 'plan_list', 'html_content', 'subscribe_button_text', 'order'
    ]
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    success_message = 'Subscription plan successfully added to plan list'
    template_name = 'subscriptions/plan_list_detail_create.html'

    def get_context_data(self, **kwargs):
        """Extend context to include the parent PlanList object."""
        context = super().get_context_data(**kwargs)

        context['plan_list'] = get_object_or_404(
            models.PlanList, id=self.kwargs.get('plan_list_id', None)
        )

        return context

    def get_success_url(self):
        return reverse_lazy(
            'dfs_plan_list_detail_list',
            kwargs={'plan_list_id': self.kwargs['plan_list_id']},
        )


class PlanListDetailUpdateView(
        PermissionRequiredMixin, SuccessMessageMixin, abstract.UpdateView
):
    """View to update the details of a plan list detail."""
    model = models.PlanListDetail
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    fields = [
        'plan', 'plan_list', 'html_content', 'subscribe_button_text', 'order'
    ]
    success_message = 'Plan list details successfully updated'
    pk_url_kwarg = 'plan_list_detail_id'
    template_name = 'subscriptions/plan_list_detail_update.html'

    def get_context_data(self, **kwargs):
        """Extend context to include the parent PlanList object."""
        context = super().get_context_data(**kwargs)

        context['plan_list'] = get_object_or_404(
            models.PlanList, id=self.kwargs.get('plan_list_id', None)
        )

        return context

    def get_success_url(self):
        return reverse_lazy(
            'dfs_plan_list_detail_list',
            kwargs={'plan_list_id': self.kwargs['plan_list_id']},
        )


class PlanListDetailDeleteView(PermissionRequiredMixin, abstract.DeleteView):
    """View to delete a plan list detail.

        View is extended to handle additional attributes noted below.

        Attributes:
            success_message (str): Message to display on successful deletion.
            success_url (str): URL to redirect to on successful deletion.
    """
    model = models.PlanListDetail
    permission_required = 'subscriptions.subscriptions'
    raise_exception = True
    context_object_name = 'plan_list_detail'
    pk_url_kwarg = 'plan_list_detail_id'
    success_message = 'Subscription plan successfully removed from plan list'
    template_name = 'subscriptions/plan_list_detail_delete.html'

    def get_context_data(self, **kwargs):
        """Extend context to include the parent PlanList object."""
        context = super().get_context_data(**kwargs)

        context['plan_list'] = get_object_or_404(
            models.PlanList, id=self.kwargs.get('plan_list_id', None)
        )

        return context

    def delete(self, request, *args, **kwargs):
        """Override delete to allow success message to be added."""
        messages.success(self.request, self.success_message)
        return super(PlanListDetailDeleteView, self).delete(
            request, *args, **kwargs
        )

    def get_success_url(self):
        return reverse_lazy(
            'dfs_plan_list_detail_list',
            kwargs={'plan_list_id': self.kwargs['plan_list_id']},
        )


# Subscribe Views
# -----------------------------------------------------------------------------
class SubscribeList(abstract.TemplateView):
    """Detail view of the first active PlanList instance.

        View is designed to be the user-facing subscription list and
        customizable through the PlanList and PlanListDetail models.
    """
    context_object_name = 'plan_list'
    template_name = 'subscriptions/subscribe_list.html'

    def get(self, request, *args, **kwargs):
        """Ensures content is available to display, then returns page."""
        # Get the appropriate plan list
        plan_list = models.PlanList.objects.filter(active=True).first()

        # Retrieve the plan details for template display
        details = models.PlanListDetail.objects.filter(
            plan_list=plan_list, plan__costs__isnull=False
        ).order_by('order')

        if plan_list:
            response = TemplateResponse(
                request,
                self.template_name,
                self.get_context_data(plan_list=plan_list, details=details)
            )

            return response

        return HttpResponseNotFound('No subscription plans are available')

    def get_context_data(self, **kwargs):
        """Extend context to include the parent PlanList object."""
        context = super().get_context_data(**kwargs)

        context['plan_list'] = kwargs['plan_list']
        context['details'] = kwargs['details']

        return context


class SubscribeView(LoginRequiredMixin, abstract.TemplateView):
    """View to handle all aspects of the subscribing process.

        This view will need to be subclassed and some methods
        overridden to implement the payment solution.

        Additionally, this view is extended from a TemplateView with
        the additional attributes noted below.

        Attributes:
            payment_form (obj): Django Form to handle subscription
                payment.
            subscription_plan (obj): A SubscriptionPlan instance. Will
                be set by methods during processing.
            success_url (str): URL to redirect to on successful
                creation.
            template_preview (str): Path to HTML template for the
                preview view.
            template_confirmation (str): Path to HTML template for the
                confirmation view.

        Notes:
            View only accessible via POST requests. The request must
            include an ID to a SubscriptionPlan +/- associated PlanCost
            instance (if past the preview view).
    """
    confirmation = False
    payment_form = forms.PaymentForm
    subscription_plan = None
    success_url = 'dfs_subscribe_thank_you'
    template_preview = 'subscriptions/subscribe_preview.html'
    template_confirmation = 'subscriptions/subscribe_confirmation.html'

    def get_object(self):
        """Gets the subscription plan object."""
        return get_object_or_404(
            models.SubscriptionPlan, id=self.request.POST.get('plan_id', None)
        )

    def get_context_data(self, **kwargs):
        """Overriding get_context_data to add additional context."""
        context = super(SubscribeView, self).get_context_data(**kwargs)

        # Whether this is a preview or confirmation step
        context['confirmation'] = self.confirmation

        # The plan instance to use for generating plan details
        context['plan'] = self.subscription_plan

        return context

    def get_template_names(self):
        """Returns the proper template name based on payment stage."""
        conf_templates = [self.template_confirmation]
        prev_templates = [self.template_preview]

        return conf_templates if self.confirmation else prev_templates

    def get_success_url(self, **kwargs):
        """Returns the success URL."""
        return reverse_lazy(self.success_url, kwargs=kwargs)

    def get(self, request, *args, **kwargs):
        """Returns 404 error as this method is not implemented."""
        return HttpResponseNotAllowed(['POST'])

    def post(self, request, *args, **kwargs):  # pylint: disable=unused-argument
        """Handles all POST requests to the SubscribeView.

            The 'action' POST argument is used to determine which
            context to render.

            Notes:
                The ``action`` POST parameter determines what stage to
                progress view to. ``None`` directs to preview
                processing, ``confirm`` directs to confirmation
                processing, and ``process`` directs to payment and
                subscription processing.
        """
        # Get the subscription plan for this POST
        self.subscription_plan = self.get_object()

        # Determine POST action and direct to proper function
        post_action = request.POST.get('action', None)

        if post_action == 'confirm':
            return self.render_confirmation(request)

        if post_action == 'process':
            return self.process_subscription(request)

        # No action - assumes payment details need to be collected
        return self.render_preview(request)

    def render_preview(self, request, **kwargs):
        """Renders preview of subscription and collect payment details."""
        self.confirmation = False
        context = self.get_context_data(**kwargs)

        # Forms to collect subscription details
        if 'error' in kwargs:
            plan_cost_form = forms.SubscriptionPlanCostForm(
                request.POST, subscription_plan=self.subscription_plan
            )
            payment_form = self.payment_form(request.POST)
        else:
            plan_cost_form = forms.SubscriptionPlanCostForm(
                initial=request.POST, subscription_plan=self.subscription_plan
            )
            payment_form = self.payment_form(initial=request.POST)

        context['plan_cost_form'] = plan_cost_form
        context['payment_form'] = payment_form

        return self.render_to_response(context)

    def render_confirmation(self, request, **kwargs):
        """Renders a confirmation page before processing payment.

            If forms are invalid will return to preview view for user
            to correct errors.
        """
        # Retrive form details
        plan_cost_form = forms.SubscriptionPlanCostForm(
            request.POST, subscription_plan=self.subscription_plan
        )
        payment_form = self.payment_form(request.POST)

        # Validate form submission
        if all([payment_form.is_valid(), plan_cost_form.is_valid()]):
            self.confirmation = True
            context = self.get_context_data(**kwargs)

            # Forms to process payment (hidden to prevent editing)
            context['plan_cost_form'] = self.hide_form(plan_cost_form)
            context['payment_form'] = self.hide_form(payment_form)

            # Add the PlanCost instance to context for use in template
            context['plan_cost'] = plan_cost_form.cleaned_data['plan_cost']

            return self.render_to_response(context)

        # Invalid form submission - render preview again
        kwargs['error'] = True
        return self.render_preview(request, **kwargs)

    def process_subscription(self, request, **kwargs):
        """Moves forward with payment & subscription processing.

            If forms are invalid will move back to confirmation page
            for user to correct errors.
        """
        # Validate payment details again incase anything changed
        plan_cost_form = forms.SubscriptionPlanCostForm(
            request.POST, subscription_plan=self.subscription_plan
        )
        payment_form = self.payment_form(request.POST)

        if all([payment_form.is_valid(), plan_cost_form.is_valid()]):
            # Attempt to process payment
            payment_transaction = self.process_payment(
                payment_form=payment_form,
                plan_cost_form=plan_cost_form,
            )

            if payment_transaction:
                # Payment successful - can handle subscription processing
                subscription = self.setup_subscription(
                    request.user, plan_cost_form.cleaned_data['plan_cost']
                )

                # Record the transaction details
                transaction = self.record_transaction(
                    subscription,
                    self.retrieve_transaction_date(payment_transaction)
                )

                return HttpResponseRedirect(
                    self.get_success_url(transaction_id=transaction.id)
                )

            # Payment unsuccessful, add message for confirmation page
            messages.error(request, 'Error processing payment')

        # Invalid form submission/payment - render preview again
        return self.render_confirmation(request, **kwargs)

    def hide_form(self, form):
        """Replaces form widgets with hidden inputs.

            Parameters:
                form (obj): A form instance.

            Returns:
                obj: The modified form instance.
        """
        for _, field in form.fields.items():
            field.widget = HiddenInput()

        return form

    def process_payment(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Processes payment and confirms if payment is accepted.

            This method needs to be overriden in a project to handle
            payment processing with the appropriate payment provider.

            Can return value that evalutes to ``True`` to indicate
            payment success and any value that evalutes to ``False`` to
            indicate payment error.
        """
        return True

    def setup_subscription(self, request_user, plan_cost):
        """Adds subscription to user and adds them to required group.

            Parameters:
                request_user (obj): A Django user instance.
                plan_cost (obj): A PlanCost instance.

            Returns:
                obj: The newly created UserSubscription instance.
        """
        current_date = timezone.now()

        # Add subscription plan to user
        subscription = models.UserSubscription.objects.create(
            user=request_user,
            subscription=plan_cost,
            date_billing_start=current_date,
            date_billing_end=None,
            date_billing_last=current_date,
            date_billing_next=plan_cost.next_billing_datetime(current_date),
            active=True,
            cancelled=False,
        )

        # Add user to the proper group
        try:
            group = self.subscription_plan.group
            group.user_set.add(request_user)
        except AttributeError:
            # No group available to add user to
            pass

        return subscription

    def retrieve_transaction_date(self, payment):  # pylint: disable=unused-argument
        """Returns the transaction date from provided payment details.

            Method should be overriden to accomodate the implemented
            payment processing if a more accurate datetime is required.


            Returns
                obj: The current datetime.
        """
        return timezone.now()

    def record_transaction(self, subscription, transaction_date=None):
        """Records transaction details in SubscriptionTransaction.

            Parameters:
                subscription (obj): A UserSubscription object.
                transaction_date (obj): A DateTime object of when
                    payment occurred (defaults to current datetime if
                    none provided).

            Returns:
                obj: The created SubscriptionTransaction instance.
        """
        if transaction_date is None:
            transaction_date = timezone.now()

        return models.SubscriptionTransaction.objects.create(
            user=subscription.user,
            subscription=subscription.subscription,
            date_transaction=transaction_date,
            amount=subscription.subscription.cost,
        )


class SubscribeUserList(LoginRequiredMixin, abstract.ListView):
    """List of all a user's subscriptions."""
    model = models.UserSubscription
    context_object_name = 'subscriptions'
    template_name = 'subscriptions/subscribe_user_list.html'

    def get_queryset(self):
        """Overrides get_queryset to restrict list to logged in user."""
        return self.model.objects.filter(user=self.request.user, active=True)


class SubscribeThankYouView(LoginRequiredMixin, abstract.TemplateView):
    """A thank you page and summary for a new subscription."""
    template_name = 'subscriptions/subscribe_thank_you.html'
    context_object_name = 'transaction'

    def get_object(self):
        """Returns the provided transaction instance."""
        try:
            return models.SubscriptionTransaction.objects.get(
                id=self.kwargs['transaction_id'],
                user=self.request.user,
            )
        except models.SubscriptionTransaction.DoesNotExist:
            return None

    def get_context_data(self, **kwargs):
        """Overriding get_context_data to add additional context."""
        context = super(SubscribeThankYouView, self).get_context_data(**kwargs)

        # Adds the context object
        context[self.context_object_name] = self.get_object()

        return context


class SubscribeCancelView(LoginRequiredMixin, abstract.DetailView):
    """View to handle cancelling of subscription.

        View is extended to handle additional attributes noted below.

        Attributes:
            success_message (str): Message to display on successful creation.
            success_url (str): URL to redirect to on successful creation.
    """
    model = models.UserSubscription
    context_object_name = 'subscription'
    pk_url_kwarg = 'subscription_id'
    success_message = 'Subscription successfully cancelled'
    success_url = 'dfs_subscribe_user_list'
    template_name = 'subscriptions/subscribe_cancel.html'

    def get_object(self, queryset=None):
        """Overrides get_object to restrict to logged in user."""
        return get_object_or_404(
            self.model,
            user=self.request.user,
            id=self.kwargs['subscription_id'],
        )

    def get_success_url(self):
        """Returns the success URL."""
        return reverse_lazy(self.success_url)

    def post(self, request, *args, **kwargs):  # pylint: disable=unused-argument
        """Updates a subscription's details to cancel it."""
        subscription = self.get_object()
        subscription.date_billing_end = copy(subscription.date_billing_next)
        subscription.date_billing_next = None
        subscription.cancelled = True
        subscription.save()

        messages.success(self.request, self.success_message)

        return HttpResponseRedirect(self.get_success_url())



"""Part 1 of migration to switch recurrence_unit to CharField."""
# pylint: disable=missing-docstring, invalid-name
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('subscriptions', '0003_update_plan_list_detail'),
    ]

    operations = [
        migrations.RenameField(
            model_name='plancost',
            old_name='recurrence_unit',
            new_name='old_recurrence_unit',
        ),
        migrations.AddField(
            model_name='plancost',
            name='recurrence_unit',
            field=models.CharField(
                choices=[
                    ('0', 'once'),
                    ('1', 'second'),
                    ('2', 'minute'),
                    ('3', 'hour'),
                    ('4', 'day'),
                    ('5', 'week'),
                    ('6', 'month'),
                    ('7', 'year')
                ],
                default='m',
                max_length=1
            ),
        ),
    ]



# Generated by Django 3.0.3 on 2020-02-17 01:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('subscriptions', '0005_update_recurrence_unit_default'),
    ]

    operations = [
        migrations.AddField(
            model_name='plancost',
            name='slug',
            field=models.SlugField(
                blank=True,
                help_text='slug to reference these cost details',
                max_length=128,
                null=True,
                unique=True,
            ),
        ),
        migrations.AddField(
            model_name='subscriptionplan',
            name='slug',
            field=models.SlugField(
                blank=True,
                help_text='slug to reference the subscription plan',
                max_length=128,
                null=True,
                unique=True,
            ),
        ),
        migrations.AddField(
            model_name='planlist',
            name='slug',
            field=models.SlugField(
                blank=True,
                help_text='slug to reference the subscription plan list',
                max_length=128,
                null=True,
                unique=True,
            ),
        ),
    ]



"""Module to handle details involving currency formating."""
from decimal import Decimal, ROUND_HALF_UP


# Convenience values for sign positions
SIGN_PARANTHESES = 0
SIGN_PRECEDE_VALUE_SYMBOL = 1
SIGN_FOLLOW_VALUE_SYMBOL = 2
SIGN_PRECEDE_VALUE = 3
SIGN_FOLLOW_VALUE = 4

CURRENCY = {
    'de_de': {
        'currency_symbol': '€',
        'int_curr_symbol': 'EUR',
        'p_cs_precedes': False,
        'n_cs_precedes': False,
        'p_sep_by_space': True,
        'n_sep_by_space': True,
        'mon_decimal_point': ',',
        'mon_thousands_sep': '.',
        'mon_grouping': 3,
        'frac_digits': 2,
        'int_frac_digits': 2,
        'positive_sign': '',
        'negative_sign': '-',
        'p_sign_posn': SIGN_PRECEDE_VALUE_SYMBOL,
        'n_sign_posn': SIGN_PARANTHESES,
    },
    'en_au': {
        'currency_symbol': '$',
        'int_curr_symbol': 'AUD',
        'p_cs_precedes': True,
        'n_cs_precedes': True,
        'p_sep_by_space': False,
        'n_sep_by_space': False,
        'mon_decimal_point': '.',
        'mon_thousands_sep': ',',
        'mon_grouping': 3,
        'frac_digits': 2,
        'int_frac_digits': 2,
        'positive_sign': '',
        'negative_sign': '-',
        'p_sign_posn': SIGN_PRECEDE_VALUE,
        'n_sign_posn': SIGN_PRECEDE_VALUE_SYMBOL,
    },
    'en_ca': {
        'currency_symbol': '$',
        'int_curr_symbol': 'CAD',
        'p_cs_precedes': True,
        'n_cs_precedes': True,
        'p_sep_by_space': False,
        'n_sep_by_space': False,
        'mon_decimal_point': '.',
        'mon_thousands_sep': ',',
        'mon_grouping': 3,
        'frac_digits': 2,
        'int_frac_digits': 2,
        'positive_sign': '',
        'negative_sign': '-',
        'p_sign_posn': SIGN_PRECEDE_VALUE,
        'n_sign_posn': SIGN_PRECEDE_VALUE,
    },
    'en_us': {
        'currency_symbol': '$',
        'int_curr_symbol': 'USD',
        'p_cs_precedes': True,
        'n_cs_precedes': True,
        'p_sep_by_space': False,
        'n_sep_by_space': False,
        'mon_decimal_point': '.',
        'mon_thousands_sep': ',',
        'mon_grouping': 3,
        'frac_digits': 2,
        'int_frac_digits': 2,
        'positive_sign': '',
        'negative_sign': '-',
        'p_sign_posn': SIGN_PRECEDE_VALUE,
        'n_sign_posn': SIGN_PARANTHESES,
    },
    'fa_ir': {
        'currency_symbol': '﷼',
        'int_curr_symbol': 'IRR',
        'p_cs_precedes': False,
        'n_cs_precedes': False,
        'p_sep_by_space': True,
        'n_sep_by_space': True,
        'mon_decimal_point': '.',
        'mon_thousands_sep': ',',
        'mon_grouping': 3,
        'frac_digits': 0,
        'int_frac_digits': 0,
        'positive_sign': '',
        'negative_sign': '-',
        'p_sign_posn': SIGN_PRECEDE_VALUE_SYMBOL,
        'n_sign_posn': SIGN_PRECEDE_VALUE_SYMBOL
    },
    'fr_ca': {
        'currency_symbol': '$',
        'int_curr_symbol': 'CAD',
        'p_cs_precedes': False,
        'n_cs_precedes': False,
        'p_sep_by_space': True,
        'n_sep_by_space': True,
        'mon_decimal_point': ',',
        'mon_thousands_sep': '\xa0',
        'mon_grouping': 3,
        'frac_digits': 2,
        'int_frac_digits': 2,
        'positive_sign': '',
        'negative_sign': '-',
        'p_sign_posn': SIGN_PRECEDE_VALUE_SYMBOL,
        'n_sign_posn': SIGN_PARANTHESES,
    },
    'pl_pl': {
        'currency_symbol': 'zł',
        'int_curr_symbol': 'PLN',
        'p_cs_precedes': False,
        'n_cs_precedes': False,
        'p_sep_by_space': True,
        'n_sep_by_space': True,
        'mon_decimal_point': ',',
        'mon_thousands_sep': '.',
        'mon_grouping': 3,
        'frac_digits': 2,
        'int_frac_digits': 2,
        'positive_sign': '',
        'negative_sign': '-',
        'p_sign_posn': SIGN_PRECEDE_VALUE_SYMBOL,
        'n_sign_posn': SIGN_PARANTHESES,
    },
    'fr_fr': {
        'currency_symbol': '€',
        'int_curr_symbol': 'EUR',
        'p_cs_precedes': False,
        'n_cs_precedes': False,
        'p_sep_by_space': True,
        'n_sep_by_space': True,
        'mon_decimal_point': ',',
        'mon_thousands_sep': '.',
        'mon_grouping': 3,
        'frac_digits': 2,
        'int_frac_digits': 2,
        'positive_sign': '',
        'negative_sign': '-',
        'p_sign_posn': SIGN_PRECEDE_VALUE,
        'n_sign_posn': SIGN_PRECEDE_VALUE,
    },
    'fr_ch': {
        'currency_symbol': 'CHF',
        'int_curr_symbol': 'CHF',
        'p_cs_precedes': False,
        'n_cs_precedes': False,
        'p_sep_by_space': True,
        'n_sep_by_space': True,
        'mon_decimal_point': '.',
        'mon_thousands_sep': "'",
        'mon_grouping': 3,
        'frac_digits': 2,
        'int_frac_digits': 2,
        'positive_sign': '',
        'negative_sign': '-',
        'p_sign_posn': SIGN_PRECEDE_VALUE,
        'n_sign_posn': SIGN_PRECEDE_VALUE,
    },
    'it_it': {
        'currency_symbol': '€',
        'int_curr_symbol': 'EUR',
        'p_cs_precedes': False,
        'n_cs_precedes': False,
        'p_sep_by_space': True,
        'n_sep_by_space': True,
        'mon_decimal_point': ',',
        'mon_thousands_sep': ".",
        'mon_grouping': 3,
        'frac_digits': 2,
        'int_frac_digits': 2,
        'positive_sign': '',
        'negative_sign': '-',
        'p_sign_posn': SIGN_PRECEDE_VALUE,
        'n_sign_posn': SIGN_PRECEDE_VALUE,
    },
    'pt_br': {
        'currency_symbol': 'R$',
        'int_curr_symbol': 'BRL',
        'p_cs_precedes': False,
        'n_cs_precedes': False,
        'p_sep_by_space': True,
        'n_sep_by_space': True,
        'mon_decimal_point': ',',
        'mon_thousands_sep': ".",
        'mon_grouping': 3,
        'frac_digits': 2,
        'int_frac_digits': 2,
        'positive_sign': '',
        'negative_sign': '-',
        'p_sign_posn': SIGN_PRECEDE_VALUE,
        'n_sign_posn': SIGN_PRECEDE_VALUE,
    },
    'en_in': {
        'currency_symbol': '₹',
        'int_curr_symbol': 'INR',
        'p_cs_precedes': True,
        'n_cs_precedes': True,
        'p_sep_by_space': False,
        'n_sep_by_space': False,
        'mon_decimal_point': '.',
        'mon_thousands_sep': ',',
        'mon_grouping': 3,
        'frac_digits': 2,
        'int_frac_digits': 2,
        'positive_sign': '',
        'negative_sign': '-',
        'p_sign_posn': SIGN_PRECEDE_VALUE,
        'n_sign_posn': SIGN_PARANTHESES,
    },
}


class Currency():
    """Defines and outputs formatted currency strings.

        Parameters:
            currency_locale (str or dict): a currency locale string or
                a dictionary defining custom currency formating
                conventions.
    """
    def __init__(self, currency_locale):
        self.international = False
        self.locale = None
        self.conventions = self._assign_currency_conventions(currency_locale)

    def _assign_currency_conventions(self, currency_locale):
        """Assigns currency conventions based on specified locale.

            If a dictionary is provided, the user is
            manually specifying the formating conventions. If a string
            is provided, the user is using a predefined currency locale.

            Parameters:
                currency_locale (str or dict): the currency locale to
                    use for formating conventions.

            Returns:
                dict: dictionary of formating conventions.
        """
        # If dictionary is passed, user is specifying their own values
        if isinstance(currency_locale, dict):
            self.locale = 'custom'

            return {
                'currency_symbol': currency_locale.get('currency_symbol', ''),
                'int_curr_symbol': currency_locale.get('int_curr_symbol', ''),
                'p_cs_precedes': currency_locale.get('p_cs_precedes', True),
                'n_cs_precedes': currency_locale.get('n_cs_precedes', True),
                'p_sep_by_space': currency_locale.get('p_sep_by_space', False),
                'n_sep_by_space': currency_locale.get('n_sep_by_space', False),
                'mon_decimal_point': currency_locale.get('mon_decimal_point', '.'),
                'mon_thousands_sep': currency_locale.get('mon_thousands_sep', ''),
                'mon_grouping': currency_locale.get('mon_grouping', 3),
                'frac_digits': currency_locale.get('frac_digits', 2),
                'int_frac_digits': currency_locale.get('int_frac_digits', 2),
                'positive_sign': currency_locale.get('positive_sign', ''),
                'negative_sign': currency_locale.get('negative_sign', ''),
                'p_sign_posn': currency_locale.get('p_sign_posn', ''),
                'n_sign_posn': currency_locale.get('n_sign_posn', ''),
            }

        # Otherwise, they have specified an existing locale
        self.locale = currency_locale.lower()

        return CURRENCY[self.locale]

    def _determine_frac_digits(self):
        """Determines number of fractional digits to round to.

            Returns:
                int: the number of fractional digits for currency display.
        """
        if self.international:
            return self.conventions['int_frac_digits']

        return self.conventions['frac_digits']

    def _split_value(self, value):
        """Splits provided value into whole and fractional parts.

            Parameters:
                value (dec): the value to split into components.

            Returns:
                tuple: the whole number and fractional number
                    components as strings.
        """
        # Determine number of fractional digits to use for this value
        digits = self._determine_frac_digits()

        # Round to required number of digits:
        #   - uses ROUND_HALF_UP, to give the most intuitive result to a
        #     typical user
        #   - uses absolute value as negative signs will be applied later
        decimal_str = abs(value).quantize(
            Decimal(10) ** -digits, rounding=ROUND_HALF_UP,
        )

        # Split decimal into whole and fractions
        try:
            num_whole, num_frac = str(decimal_str).split('.')
        except ValueError:
            num_whole, num_frac = str(decimal_str), ''

        return num_whole, num_frac

    def _group_whole_num(self, num_whole):
        """Formats whole number into appropriate groups.

            Parameters:
                num_whole (str): the whole number portion of the value.

            Returns:
                str: the whole number with appropriate grouping applied.
        """
        remaining = num_whole
        groups = []

        while remaining:
            groups.append(remaining[-self.conventions['mon_grouping']:])
            remaining = remaining[:-self.conventions['mon_grouping']]

        groups.reverse()
        grouped_num_whole = self.conventions['mon_thousands_sep'].join(groups)

        return grouped_num_whole

    def _format_value(self, num_whole, num_frac):
        """Returns formatted value with appropriate decimal seperator.

            Parameters:
                num_whole (str): the whole number portion of the value.
                num_frac (str): the fractional portion of the value.

            Returns:
                str: combined value, separated by the decimal separator.
        """
        frac_digits = self._determine_frac_digits()

        # Determines decimal separator (only required if currency uses fractions)
        if frac_digits > 0:
            dec_separator = self.conventions['mon_decimal_point']
        else:
            dec_separator = ''

        # Assemble and return the formatted value
        return '{}{}{}'.format(
            num_whole, dec_separator, num_frac
        )

    def _determine_symbol_details(self, negative_value):
        """Determines positioning of required symbols.

            Parameters:
                negative_value (bool): whether this is a negative
                    value or not.

            Returns:
                obj: Currency symbol and positioning details.
        """
        # Determine which symbol to use
        if self.international:
            symbol = self.conventions['int_curr_symbol']
        else:
            symbol = self.conventions['currency_symbol']

        # If value less than zero, assign negative symbol details
        if negative_value:
            return {
                'symbol': symbol,
                'precedes': self.conventions['n_cs_precedes'],
                'separated': ' ' if self.conventions['n_sep_by_space'] else '',
                'sign': self.conventions['negative_sign'],
                'sign_position': self.conventions['n_sign_posn'],
            }

        # Otherwise return positive symbol details
        return {
            'symbol': symbol,
            'precedes': self.conventions['p_cs_precedes'],
            'separated': ' ' if self.conventions['p_sep_by_space'] else '',
            'sign': self.conventions['positive_sign'],
            'sign_position': self.conventions['p_sign_posn'],
        }

    def add_symbols(self, value, negative_value):
        """Adds currency and positive/negative symbols to the value.

            Parameters:
                value (str): the formatted value to add
                    symbols to.
                negative_value (bool): whether this is a negative
                    value or not.

            Returns:
                str: the final value formatted as a currency value.
        """
        # '<' and '>' are used as markers of number start and end
        placeholder_value = '<{}>'.format(value)

        # Determine details of additional formatting
        symbol = self._determine_symbol_details(negative_value)

        if symbol['precedes']:
            placeholder_value = '{}{}{}'.format(
                symbol['symbol'], symbol['separated'], placeholder_value
            )
        else:
            placeholder_value = '{}{}{}'.format(
                placeholder_value, symbol['separated'], symbol['symbol']
            )

        # Insert the proper sign for positive/negative values
        if symbol['sign_position'] == 0:
            placeholder_value = '({})'.format(placeholder_value)
        elif symbol['sign_position'] == 1:
            placeholder_value = '{}{}'.format(symbol['sign'], placeholder_value)
        elif symbol['sign_position'] == 2:
            placeholder_value = '{}{}'.format(placeholder_value, symbol['sign'])
        elif symbol['sign_position'] == 3:
            placeholder_value = placeholder_value.replace('<', symbol['sign'])
        elif symbol['sign_position'] == 4:
            placeholder_value = placeholder_value.replace('>', symbol['sign'])
        else:
            placeholder_value = '{}{}'.format(symbol['sign'], placeholder_value)

        # Remove the placeholders for number start/end
        formatted_currency = (
            placeholder_value.replace('<', '').replace('>', '')
        )

        return formatted_currency

    def format_currency(self, value, international=False):
        """Returns the provided value in the proper currency format.

            Parameters:
                value (dec): The decimal to represent as a currency.
                international (bool): Whether this should follow
                    international formatting or not.

            Returns:
                str: The formatted currency value.
        """
        # Assign international status for currency format
        self.international = international

        # Convert value to decimal for future operations
        value = Decimal(value)

        # Split value into whole and fractional parts
        num_whole, num_frac = self._split_value(value)

        # Apply any grouping to the whole number component
        grouped_num_whole = self._group_whole_num(num_whole)

        # Rejoin whole and fractional numbers with decimal separator
        formatted_value = self._format_value(grouped_num_whole, num_frac)

        # Add currency and positive/negative symbols
        formatted_currency = self.add_symbols(formatted_value, bool(value < 0))

        return formatted_currency



"""Utility/helper functions for Django Flexible Subscriptions."""
from django.db.models import Q
from django.utils import timezone

from subscriptions import models


class Manager():
    """Manager object to help manage subscriptions & billing."""

    def process_subscriptions(self):
        """Calls all required subscription processing functions."""
        current = timezone.now()

        # Handle expired subscriptions
        expired_subscriptions = models.UserSubscription.objects.filter(
            Q(active=True) & Q(cancelled=False)
            & Q(date_billing_end__lte=current)
        )

        for subscription in expired_subscriptions:
            self.process_expired(subscription)

        # Handle new subscriptions
        new_subscriptions = models.UserSubscription.objects.filter(
            Q(active=False) & Q(cancelled=False)
            & Q(date_billing_start__lte=current)
        )

        for subscription in new_subscriptions:
            self.process_new(subscription)

        # Handle subscriptions with billing due
        due_subscriptions = models.UserSubscription.objects.filter(
            Q(active=True) & Q(cancelled=False)
            & Q(date_billing_next__lte=current)
        )

        for subscription in due_subscriptions:
            self.process_due(subscription)

    def process_expired(self, subscription):
        """Handles processing of expired/cancelled subscriptions.

            Parameters:
                subscription (obj): A UserSubscription instance.
        """
        # Get all user subscriptions
        user = subscription.user
        user_subscriptions = user.subscriptions.all()
        subscription_group = subscription.subscription.plan.group
        group_matches = 0

        # Check if there is another subscription for this group
        for user_subscription in user_subscriptions:
            if user_subscription.subscription.plan.group == subscription_group:
                group_matches += 1

        # If no other subscription, can remove user from group
        if group_matches < 2:
            subscription_group.user_set.remove(user)

        # Update this specific UserSubscription instance
        subscription.active = False
        subscription.cancelled = True
        subscription.save()

        self.notify_expired(subscription)

    def process_new(self, subscription):
        """Handles processing of a new subscription.

            Parameters:
                subscription (obj): A UserSubscription instance.
        """
        user = subscription.user
        cost = subscription.subscription
        plan = cost.plan

        payment_transaction = self.process_payment(user=user, cost=cost)

        if payment_transaction:
            # Add user to the proper group
            try:
                plan.group.user_set.add(user)
            except AttributeError:
                # No group available to add user to
                pass

            # Update subscription details
            current = timezone.now()
            next_billing = cost.next_billing_datetime(
                subscription.date_billing_start
            )
            subscription.date_billing_last = current
            subscription.date_billing_next = next_billing
            subscription.active = True
            subscription.save()

            # Record the transaction details
            self.record_transaction(
                subscription,
                self.retrieve_transaction_date(payment_transaction)
            )

            # Send notifications
            self.notify_new(subscription)

    def process_due(self, subscription):
        """Handles processing of a due subscription.

            Parameters:
                subscription (obj): A UserSubscription instance.
        """
        user = subscription.user
        cost = subscription.subscription

        payment_transaction = self.process_payment(user=user, cost=cost)

        if payment_transaction:
            # Update subscription details
            current = timezone.now()
            next_billing = cost.next_billing_datetime(
                subscription.date_billing_next
            )
            subscription.date_billing_last = current
            subscription.date_billing_next = next_billing
            subscription.save()

            # Record the transaction details
            self.record_transaction(
                subscription,
                self.retrieve_transaction_date(payment_transaction)
            )

    def process_payment(self, *args, **kwargs):  # pylint: disable=unused-argument, no-self-use
        """Processes payment and confirms if payment is accepted.

            This method needs to be overriden in a project to handle
            payment processing with the appropriate payment provider.

            Can return value that evalutes to ``True`` to indicate
            payment success and any value that evalutes to ``False`` to
            indicate payment error.
        """
        return True

    def retrieve_transaction_date(self, payment):  # pylint: disable=unused-argument, no-self-use
        """Returns the transaction date from provided payment details.

            Method should be overriden to accomodate the implemented
            payment processing if a more accurate datetime is required.


            Returns
                obj: The current datetime.
        """
        return timezone.now()

    @staticmethod
    def record_transaction(subscription, transaction_date=None):
        """Records transaction details in SubscriptionTransaction.

            Parameters:
                subscription (obj): A UserSubscription object.
                transaction_date (obj): A DateTime object of when
                    payment occurred (defaults to current datetime if
                    none provided).

            Returns:
                obj: The created SubscriptionTransaction instance.
        """
        if transaction_date is None:
            transaction_date = timezone.now()

        return models.SubscriptionTransaction.objects.create(
            user=subscription.user,
            subscription=subscription.subscription,
            date_transaction=transaction_date,
            amount=subscription.subscription.cost,
        )

    def notify_expired(self, subscription):
        """Sends notification of expired subscription.

            Parameters:
                subscription (obj): A UserSubscription instance.
        """

    def notify_new(self, subscription):
        """Sends notification of newly active subscription

            Parameters:
                subscription (obj): A UserSubscription instance.
        """

    def notify_payment_error(self, subscription):
        """Sends notification of a payment error

            Parameters:
                subscription (obj): A UserSubscription instance.
        """

    def notify_payment_success(self, subscription):
        """Sends notifiation of a payment success

            Parameters:
                subscription (obj): A UserSubscription instance.
        """



"""PyPI setup script for the django-flexible-subscriptions package."""
from setuptools import find_packages, setup

from subscriptions import __version__

with open('README.rst', 'r') as readme_file:
    LONG_DESCRIPTION = readme_file.read()

setup(
    name='django-flexible-subscriptions',
    version=__version__,
    url='https://github.com/studybuffalo/django-flexible-subscriptions',
    description=('A subscription and recurrent billing application for Django.'),
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/x-rst',
    author='Joshua Torrance',
    author_email='studybuffalo@gmail.com',
    keywords='Django, subscriptions, recurrent billing',
    platforms=['linux', 'windows'],
    packages=find_packages(exclude=['sandbox*', 'tests*']),
    package_data={
        'subscriptions': [
            'management/commands/*.py',
            'static/subscriptions/*.css',
            'templates/subscriptions/*.html',
            'templates/subscriptions/snippets/*.html',
        ]
    },
    project_urls={
        'Documentation': 'https://django-flexible-subscriptions.readthedocs.io/en/latest/',
        'Source code': 'https://github.com/studybuffalo/django-flexible-subscriptions',
        'Issues': 'https://github.com/studybuffalo/django-flexible-subscriptions/issues',
    },
    python_requires='>=3.5',
    install_requires=[
        'django>=2.2',
    ],
    tests_require=[
        'pytest==5.4.3',
        'pytest-cov==2.10.0',
    ],
    # See http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Other/Nonlisted Topic'
    ],
)



# pylint: disable=missing-docstring
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('subscriptions', '0002_plan_list_addition'),
    ]

    operations = [
        migrations.AddField(
            model_name='planlistdetail',
            name='order',
            field=models.PositiveIntegerField(
                default=1,
                help_text='Order to display plan in (lower numbers displayed first)'
            ),
        ),
        migrations.RemoveField(
            model_name='planlist',
            name='plans',
        ),
        migrations.AlterField(
            model_name='planlistdetail',
            name='plan',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='plan_list_details',
                to='subscriptions.SubscriptionPlan'
            ),
        ),
        migrations.AlterField(
            model_name='planlistdetail',
            name='plan_list',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='plan_list_details',
                to='subscriptions.PlanList'
            ),
        ),
    ]



"""Migration to fix incorrect default for recurrence unit."""
# pylint: disable=invalid-name
from django.db import migrations, models


class Migration(migrations.Migration):
    """Updates plancost recurrence_unit default to '6' instead of 'm'."""
    dependencies = [
        ('subscriptions', '0004_change_recurrence_unit_type_2'),
    ]

    operations = [
        migrations.AlterField(
            model_name='plancost',
            name='recurrence_unit',
            field=models.CharField(
                choices=[
                    ('0', 'once'),
                    ('1', 'second'),
                    ('2', 'minute'),
                    ('3', 'hour'),
                    ('4', 'day'),
                    ('5', 'week'),
                    ('6', 'month'),
                    ('7', 'year'),
                ],
                default='6',
                max_length=1
            ),
        ),
    ]



"""Models for the Flexible Subscriptions app."""
from datetime import timedelta
from uuid import uuid4

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.core.validators import MinValueValidator
from django.db import models
from django.utils.translation import gettext_lazy as _


# Convenience references for units for plan recurrence billing
# ----------------------------------------------------------------------------
ONCE = '0'
SECOND = '1'
MINUTE = '2'
HOUR = '3'
DAY = '4'
WEEK = '5'
MONTH = '6'
YEAR = '7'
RECURRENCE_UNIT_CHOICES = (
    (ONCE, 'once'),
    (SECOND, 'second'),
    (MINUTE, 'minute'),
    (HOUR, 'hour'),
    (DAY, 'day'),
    (WEEK, 'week'),
    (MONTH, 'month'),
    (YEAR, 'year'),
)


class PlanTag(models.Model):
    """A tag for a subscription plan."""
    tag = models.CharField(
        help_text=_('the tag name'),
        max_length=64,
        unique=True,
    )

    class Meta:
        ordering = ('tag',)

    def __str__(self):
        return self.tag


class SubscriptionPlan(models.Model):
    """Details for a subscription plan."""
    id = models.UUIDField(
        default=uuid4,
        editable=False,
        primary_key=True,
        verbose_name='ID',
    )
    plan_name = models.CharField(
        help_text=_('the name of the subscription plan'),
        max_length=128,
    )
    slug = models.SlugField(
        blank=True,
        help_text=_('slug to reference the subscription plan'),
        max_length=128,
        null=True,
        unique=True,
    )
    plan_description = models.CharField(
        blank=True,
        help_text=_('a description of the subscription plan'),
        max_length=512,
        null=True,
    )
    group = models.ForeignKey(
        Group,
        blank=True,
        help_text=_('the Django auth group for this plan'),
        null=True,
        on_delete=models.SET_NULL,
        related_name='plans',
    )
    tags = models.ManyToManyField(
        PlanTag,
        blank=True,
        help_text=_('any tags associated with this plan'),
        related_name='plans',
    )
    grace_period = models.PositiveIntegerField(
        default=0,
        help_text=_(
            'how many days after the subscription ends before the '
            'subscription expires'
        ),
    )

    class Meta:
        ordering = ('plan_name',)
        permissions = (
            ('subscriptions', 'Can interact with subscription details'),
        )

    def __str__(self):
        return self.plan_name

    def display_tags(self):
        """Displays tags as a string (truncates if more than 3)."""
        if self.tags.count() > 3:
            return '{}, ...'.format(
                ', '.join(tag.tag for tag in self.tags.all()[:3])
            )

        return ', '.join(tag.tag for tag in self.tags.all()[:3])


class PlanCost(models.Model):
    """Cost and frequency of billing for a plan."""
    id = models.UUIDField(
        default=uuid4,
        editable=False,
        primary_key=True,
        verbose_name='ID',
    )
    plan = models.ForeignKey(
        SubscriptionPlan,
        help_text=_('the subscription plan for these cost details'),
        on_delete=models.CASCADE,
        related_name='costs',
    )
    slug = models.SlugField(
        blank=True,
        help_text=_('slug to reference these cost details'),
        max_length=128,
        null=True,
        unique=True,
    )
    recurrence_period = models.PositiveSmallIntegerField(
        default=1,
        help_text=_('how often the plan is billed (per recurrence unit)'),
        validators=[MinValueValidator(1)],
    )
    recurrence_unit = models.CharField(
        choices=RECURRENCE_UNIT_CHOICES,
        default=MONTH,
        max_length=1,
    )
    cost = models.DecimalField(
        blank=True,
        decimal_places=4,
        help_text=_('the cost per recurrence of the plan'),
        max_digits=19,
        null=True,
    )

    class Meta:
        ordering = ('recurrence_unit', 'recurrence_period', 'cost',)

    @property
    def display_recurrent_unit_text(self):
        """Converts recurrence_unit integer to text."""
        conversion = {
            ONCE: 'one-time',
            SECOND: 'per second',
            MINUTE: 'per minute',
            HOUR: 'per hour',
            DAY: 'per day',
            WEEK: 'per week',
            MONTH: 'per month',
            YEAR: 'per year',
        }

        return conversion[self.recurrence_unit]

    @property
    def display_billing_frequency_text(self):
        """Generates human-readable billing frequency."""
        conversion = {
            ONCE: 'one-time',
            SECOND: {'singular': 'per second', 'plural': 'seconds'},
            MINUTE: {'singular': 'per minute', 'plural': 'minutes'},
            HOUR: {'singular': 'per hour', 'plural': 'hours'},
            DAY: {'singular': 'per day', 'plural': 'days'},
            WEEK: {'singular': 'per week', 'plural': 'weeks'},
            MONTH: {'singular': 'per month', 'plural': 'months'},
            YEAR: {'singular': 'per year', 'plural': 'years'},
        }

        if self.recurrence_unit == ONCE:
            return conversion[ONCE]

        if self.recurrence_period == 1:
            return conversion[self.recurrence_unit]['singular']

        return 'every {} {}'.format(
            self.recurrence_period, conversion[self.recurrence_unit]['plural']
        )

    def next_billing_datetime(self, current):
        """Calculates next billing date for provided datetime.

            Parameters:
                current (datetime): The current datetime to compare
                    against.

            Returns:
                datetime: The next time billing will be due.
        """
        if self.recurrence_unit == SECOND:
            delta = timedelta(seconds=self.recurrence_period)
        elif self.recurrence_unit == MINUTE:
            delta = timedelta(minutes=self.recurrence_period)
        elif self.recurrence_unit == HOUR:
            delta = timedelta(hours=self.recurrence_period)
        elif self.recurrence_unit == DAY:
            delta = timedelta(days=self.recurrence_period)
        elif self.recurrence_unit == WEEK:
            delta = timedelta(weeks=self.recurrence_period)
        elif self.recurrence_unit == MONTH:
            # Adds the average number of days per month as per:
            # http://en.wikipedia.org/wiki/Month#Julian_and_Gregorian_calendars
            # This handle any issues with months < 31 days and leap years
            delta = timedelta(
                days=30.4368 * self.recurrence_period
            )
        elif self.recurrence_unit == YEAR:
            # Adds the average number of days per year as per:
            # http://en.wikipedia.org/wiki/Year#Calendar_year
            # This handle any issues with leap years
            delta = timedelta(
                days=365.2425 * self.recurrence_period
            )
        else:
            # If no recurrence period, no next billing datetime
            return None

        return current + delta


class UserSubscription(models.Model):
    """Details of a user's specific subscription."""
    id = models.UUIDField(
        default=uuid4,
        editable=False,
        primary_key=True,
        verbose_name='ID',
    )
    user = models.ForeignKey(
        get_user_model(),
        help_text=_('the user this subscription applies to'),
        null=True,
        on_delete=models.CASCADE,
        related_name='subscriptions',
    )
    subscription = models.ForeignKey(
        PlanCost,
        help_text=_('the plan costs and billing frequency for this user'),
        null=True,
        on_delete=models.CASCADE,
        related_name='subscriptions'
    )
    date_billing_start = models.DateTimeField(
        blank=True,
        help_text=_('the date to start billing this subscription'),
        null=True,
        verbose_name='billing start date',
    )
    date_billing_end = models.DateTimeField(
        blank=True,
        help_text=_('the date to finish billing this subscription'),
        null=True,
        verbose_name='billing start end',
    )
    date_billing_last = models.DateTimeField(
        blank=True,
        help_text=_('the last date this plan was billed'),
        null=True,
        verbose_name='last billing date',
    )
    date_billing_next = models.DateTimeField(
        blank=True,
        help_text=_('the next date billing is due'),
        null=True,
        verbose_name='next start date',
    )
    active = models.BooleanField(
        default=True,
        help_text=_('whether this subscription is active or not'),
    )
    cancelled = models.BooleanField(
        default=False,
        help_text=_('whether this subscription is cancelled or not'),
    )

    class Meta:
        ordering = ('user', 'date_billing_start',)


class SubscriptionTransaction(models.Model):
    """Details for a subscription plan billing."""
    id = models.UUIDField(
        default=uuid4,
        editable=False,
        primary_key=True,
        verbose_name='ID',
    )
    user = models.ForeignKey(
        get_user_model(),
        help_text=_('the user that this subscription was billed for'),
        null=True,
        on_delete=models.SET_NULL,
        related_name='subscription_transactions'
    )
    subscription = models.ForeignKey(
        PlanCost,
        help_text=_('the plan costs that were billed'),
        null=True,
        on_delete=models.SET_NULL,
        related_name='transactions'
    )
    date_transaction = models.DateTimeField(
        help_text=_('the datetime the transaction was billed'),
        verbose_name='transaction date',
    )
    amount = models.DecimalField(
        blank=True,
        decimal_places=4,
        help_text=_('how much was billed for the user'),
        max_digits=19,
        null=True,
    )

    class Meta:
        ordering = ('date_transaction', 'user',)


class PlanList(models.Model):
    """Model to record details of a display list of SubscriptionPlans."""
    title = models.TextField(
        blank=True,
        help_text=_('title to display on the subscription plan list page'),
        null=True,
    )
    slug = models.SlugField(
        blank=True,
        help_text=_('slug to reference the subscription plan list'),
        max_length=128,
        null=True,
        unique=True,
    )
    subtitle = models.TextField(
        blank=True,
        help_text=_('subtitle to display on the subscription plan list page'),
        null=True,
    )
    header = models.TextField(
        blank=True,
        help_text=_('header text to display on the subscription plan list page'),
        null=True,
    )
    footer = models.TextField(
        blank=True,
        help_text=_('header text to display on the subscription plan list page'),
        null=True,
    )
    active = models.BooleanField(
        default=True,
        help_text=_('whether this plan list is active or not.'),
    )

    def __str__(self):
        return self.title


class PlanListDetail(models.Model):
    """Model to add additional details to plans when part of PlanList."""
    plan = models.ForeignKey(
        SubscriptionPlan,
        on_delete=models.CASCADE,
        related_name='plan_list_details',
    )
    plan_list = models.ForeignKey(
        PlanList,
        on_delete=models.CASCADE,
        related_name='plan_list_details',
    )
    html_content = models.TextField(
        blank=True,
        help_text=_('HTML content to display for plan'),
        null=True,
    )
    subscribe_button_text = models.CharField(
        blank=True,
        default='Subscribe',
        max_length=128,
        null=True,
    )
    order = models.PositiveIntegerField(
        default=1,
        help_text=_('Order to display plan in (lower numbers displayed first)'),
    )

    def __str__(self):
        return 'Plan List {} - {}'.format(
            self.plan_list, self.plan.plan_name
        )



"""Create models to support Plan List functionality."""
# pylint: disable=missing-docstring, invalid-name
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('subscriptions', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='PlanList',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'title',
                    models.TextField(
                        blank=True,
                        help_text='title to display on the subscription plan list page',
                        null=True,
                    ),
                ),
                (
                    'subtitle',
                    models.TextField(
                        blank=True,
                        help_text='subtitle to display on the subscription plan list page',
                        null=True,
                    ),
                ),
                (
                    'header',
                    models.TextField(
                        blank=True,
                        help_text='header text to display on the subscription plan list page',
                        null=True,
                    ),
                ),
                (
                    'footer',
                    models.TextField(
                        blank=True,
                        help_text='header text to display on the subscription plan list page',
                        null=True,
                    ),
                ),
                (
                    'active',
                    models.BooleanField(
                        default=True,
                        help_text='whether this plan list is active or not.',
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name='PlanListDetail',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'html_content',
                    models.TextField(
                        blank=True,
                        help_text='HTML content to display for plan',
                        null=True,
                    ),
                ),
                (
                    'subscribe_button_text',
                    models.CharField(
                        blank=True,
                        default='Subscribe',
                        max_length=128,
                        null=True,
                    ),
                ),
                (
                    'plan',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='subscriptions.SubscriptionPlan',
                    ),
                ),
                (
                    'plan_list',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='subscriptions.PlanList',
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name='planlist',
            name='plans',
            field=models.ManyToManyField(
                blank=True,
                related_name='plan_lists',
                through='subscriptions.PlanListDetail',
                to='subscriptions.SubscriptionPlan',
            ),
        ),
    ]



"""Admin views for the Flexible Subscriptions app."""
from django.contrib import admin

from subscriptions import models
from subscriptions.conf import SETTINGS


class PlanCostInline(admin.TabularInline):
    """Inline admin class for the PlanCost model."""
    model = models.PlanCost
    fields = (
        'slug',
        'recurrence_period',
        'recurrence_unit',
        'cost',
    )
    extra = 0


class SubscriptionPlanAdmin(admin.ModelAdmin):
    """Admin class for the SubscriptionPlan model."""
    fields = (
        'plan_name',
        'slug',
        'plan_description',
        'group',
        'tags',
        'grace_period',
    )
    inlines = [PlanCostInline]
    list_display = (
        'plan_name',
        'group',
        'display_tags',
    )
    prepopulated_fields = {'slug': ('plan_name',)}


class UserSubscriptionAdmin(admin.ModelAdmin):
    """Admin class for the UserSubscription model."""
    fields = (
        'user',
        'date_billing_start',
        'date_billing_end',
        'date_billing_last',
        'date_billing_next',
        'active',
        'cancelled',
    )
    list_display = (
        'user',
        'date_billing_last',
        'date_billing_next',
        'active',
        'cancelled',
    )


class TransactionAdmin(admin.ModelAdmin):
    """Admin class for the SubscriptionTransaction model."""


if SETTINGS['enable_admin']:
    admin.site.register(models.SubscriptionPlan, SubscriptionPlanAdmin)
    admin.site.register(models.UserSubscription, UserSubscriptionAdmin)
    admin.site.register(models.SubscriptionTransaction, TransactionAdmin)



"""Application configuration file for django-flexible-subscriptions."""
from django.apps import AppConfig


class FlexibleSubscriptionsConfig(AppConfig):
    """Configuration details for django-flexible-subscriptions."""
    name = 'subscriptions'
    verbose_name = 'django-flexible-subscriptions'
